package SketchAggSensor.sensor;

/**
 * <p>Title: Aggregation Query in Sensor Network with Sketches</p>
 * <p>Description: This code is for the simulation of sensor network with sketches algorithms. It applies with aggregation queries.</p>
 * <p>Copyright: Computer Science Department, Boston University,Copyright (c) 2003</p>
 * <p>Company: Boston University</p>
 * @author Feifei Li
 * @version 1.0
 */
import java.util.*;
import SketchAggSensor.*;
import SketchAggSensor.message.*;
import SketchAggSensor.topology.*;

public class Sensor implements SensorInterface{
    public static  Random r = new Random();

    /** Specifies sensor range [0 ... MAX_READING] */
    public static final int MAX_READING = 100;

    /** How many sensor-devices are available ? */
    static final int NUM_SENSORS = 2;
    static final int LIGHT_SENSOR = 0;
    static final int TEMP_SENSOR = 1;

    /** Field to aggregate on */
    public static final int QUERY_FIELD = Sensor.LIGHT_SENSOR;

    /* Value Distributions */

    /** Randomly, uniformly selected */
    public static final int RANDOM_DIST = 0;

    /** Selected from a normal with specified mean, variance */
    public static final int NORMAL_DIST = 1;

    static final double MEAN = MAX_READING/2.0;
    static final double STD_DEV = MEAN * 0.1;

    /** Selected from a gradient */
    public static final int GEOGRAPHIC_DIST = 2;

    /** Selected from a highly-skewed distribution where 90% of values
         are in the range [middle - range/2 ... middle + range/2]
    */
    public static final int NINETY_TEN_DIST = 3;
    public static final int SUM_TEST_DIST = 4;
    static public int  middle = 50;
    static int  range = 10;

    /** Number of epochs after hearing intial request-to-aggrgeate that the
           sensor should wait before forwarding its value
       */
    public static int MAX_EPOCHS = 11;


    /** The x and y coordinates of this sensor */
    protected int x, y;
    private static final int LOC_UNKNOWN=-1;
    private int dist; //what statistic distribution are our sensor values drawn from

    /** The topology where this sensor lives */
    protected Topology top;

    /** The id of this sensor */
    protected int id;

    protected int time = 0;
    protected int numMsgs = 0;
    protected int numSent = 0;
    protected int bytesSent = 0;

    /** Readings for this sensor -- note that these are fixed over the lifetime of the device,
        generally
    */
    protected int[] reading = new int[NUM_SENSORS];
    protected boolean disabled = false;
    protected boolean noParticipate = false;

    /* Level, Parents and Children for this sensor, used in Spanning Tree */
    protected int level;
    protected Vector parents;
    protected Vector children;
    public static int LEVEL_UNKNOWN =-1;

    //public int insetIterations;
    public int queryID;
    public int sketchInsertKey;


    public boolean multipleParents = false;
    public static double NODE_FAILURE=0.0;
    public static double LINK_FAILURE=0.0;
    public static int COMPRESSION_GROUP = 20;
    public static boolean USING_DIVISOR = false;
    public static boolean useSketch = false;
    public static int numBitMaps = 20;
    public static int bitMapLength = 16;
    public static int kvalLoglog = 7;
    public boolean twoAggregator = false;
    public static int COMPRESSION_SCHEME = 1;

    /** Constructor
        @param t The topology used to radiate messages for this sensor
        @param id The id of this sensor
        @param dist The value-distribution used to generate readings for this sensor
     */
    public Sensor(Topology t, int id, int dist) {
        x = y = LOC_UNKNOWN;
        top = t;
        this.id = id;
        this.dist = dist;
        for (int i =0; i < NUM_SENSORS;i++) {
            switch(dist) {
            case RANDOM_DIST:
                reading[i] = r.nextInt(MAX_READING);
                break;
            case NORMAL_DIST:
                double p = r.nextDouble();
                reading[i] = (int)(MEAN + STD_DEV * (inverseNormal(p)));
                //		System.out.println("p = " + p + ", value = " + reading[i]);
                break;
            case GEOGRAPHIC_DIST:
                //random for now -- we'll reselect when our location is set
                reading[i] = r.nextInt(MAX_READING);
                break;
            case NINETY_TEN_DIST:
                if (r.nextDouble() < .9) {
                    reading[i] = middle;// - range + (r.nextInt(range));
                } else
                    reading[i] = r.nextInt(MAX_READING);
                break;
            case SUM_TEST_DIST:
                reading[i] = 100;
                break;
            }
        }
    }

    /** Given a probablity p, return the value of an inverse normal function at p
        centered around MEAN with variance STD_DEV^2.
    */
   double inverseNormal(double p) {
        //from http://www.hpmuseum.org/software/67pacs/67ndist.htm
        // approximation with error < 4.5x10^-4
        double c0 = 2.515517, c1 = 0.802853, c2 = 0.010328;
        double d1 = 1.432788, d2 = .189269, d3 = .001308;
        double t,y;

        if (p <= .5) {
            t = Math.sqrt(Math.log(1 / (p * p)));
        } else {
            t = Math.sqrt(Math.log(1 / ((1 - p)  * (1 - p))));
        }

        y = t - (c0 + c1*t + c2*t*t)/(1 + d1*t + d2*t*t + d3*t*t*t);
        if (p <= .5) return y;
        else return -y;
    }

    /** @return the id of this sensor */
    public int getId() {
        return id;
    }


    /** @return the current x location of this sensor */
    public int getX() {
        return x;
    }

    /** @return the current y location of this sensor */
    public int getY() {
        return y;
    }

    /** Set the x location of the sensor */
    public void setX(int x) throws InvalidLocationException{
        if (!top.validXLoc(x)) {
            throw new InvalidLocationException();
        }
        this.x = x;
        if (y != LOC_UNKNOWN && dist == GEOGRAPHIC_DIST) {
            for (int i = 0; i < NUM_SENSORS; i++) {
                reading[i] = top.getReading(x,y);
            }
        }
    }

    /** Set the y locaiton of the sensor */
    public void setY(int y) throws InvalidLocationException{
        if (!top.validYLoc(y)) {
            throw new InvalidLocationException();
        }
        this.y = y;
        if (x != LOC_UNKNOWN && dist == GEOGRAPHIC_DIST) {
            for (int i = 0; i < NUM_SENSORS; i++) {
                reading[i] = top.getReading(x,y);
            }
        }
    }

    /** Notify this sensor that it received a message */
    public void receiveMessage(Message m) {
         // numMsgs++;
    }

    /** Send a message on behalf of this sensor */
    public void sendMessage(Message m) {
        if (!m.isSpanningTreeMsg())
        {  numSent++;

           bytesSent += m.numBytes();
        }

        //allow subclasses to tweak messages before they are sent
        prepareMessage(m);

        top.radiateMessage(m, getX(), getY());
    }

    /** Allow the sensor to modify a message that is about to be sent */
    public void prepareMessage(Message m) {

    }

    /** Set the time */
    public void setTime(int time) {
        this.time = time;
    }

    /** Return the last time this sensor's setTime() method was invoked */
    public int getTime() {
        return time;
    }

    /** Return the reading of sensor-device sensorType */
    public int getReading(int sensorType) {
        return reading[sensorType];
    }

    /** Return the number of messages this sensor has heard */
    public int msgCount() {
        return numMsgs;
    }

    /** Return the number of messages this sensor has sent */
    public int sentCount() {
        return numSent;
    }

    public int sentBytes() {
        return bytesSent;
    }

    /** Disabled sensors don't forward messages for their children */
    public void setDisabled(boolean disabled) {
        this.disabled = disabled;
    }

    /** @return true iff this sensor is disabled */
    public boolean isDisabled() {
        return disabled;
    }

    /** Not participating sensors do forward messages, but don't include thier own aggs */
    public void setNoParticipate(boolean disabled) {
        this.noParticipate = true;
    }

    public int getLevel()
    {
      return level;
    }

    public void setLevel(int l)
    {
      level =l;
    }

    public Vector getParents()
    {
      return this.parents;
    }

    public Vector getChildren()
    {
      return this.children;
    }

    public void addParents(int pID)
    {
      this.parents.addElement(new Integer(pID));
    }

    public void addChildren(int childID)
    {
      this.children.addElement(new Integer(childID));
    }

    public int getAgg()
    {
      return 0;
    }

    public boolean isParent(int pID)
    {
      Enumeration e = this.parents.elements();
      while (e.hasMoreElements())
      {
        Integer id = (Integer)e.nextElement();
        if (id.intValue()==pID)
          return true;
      }
      return false;
    }

    public void setQueryID(int it)
    {
      this.queryID = it;
    }

    public void setMultipleParents(boolean p)
    {
      this.multipleParents = p;
    }

    public boolean isMultipleParents()
    {
      return this.multipleParents;
    }
}
